**Sommaire**

[[_TOC_]]

# TD #1 : Modélisation de formes géométriques

Le but de ce TD est d'illustrer les concepts d'encapsulation et d'héritage de la programmation objet, en concevant un module pour manipuler des formes géométriques avec Python. Vous commencerez par réaliser une modélisation objet du problème en définissant les classes et leurs attributs, puis implémenterez les méthodes en Python, et finalement les validerez avec des tests.

---
## Modélisation avec UML (1h30)

Les formes géométriques sont représentées par des classes, et l'héritage sera utilisé pour factoriser les propriétés communes. Nous nous limitons à un repère à deux dimensions orthonormé, avec les axes croissant vers la droite et le bas. Les coordonnées dans ce repère sont des entiers relatifs (c'est-à-dire possiblement négatifs). Dans cet espace, nous choisissons de représenter les formes suivantes :

* Les rectangles caractérisés par leur origine (`x`, `y`) et leurs dimensions (`l`, `h`).
* Les ellipses caractérisées par leur origine (`x`, `y`) et leurs rayons aux axes (`rx`, `ry`).
* Les cercles caractérisés par leur origine (`x`, `y`) et leur rayon.

<center><img src="figures/formes.svg" style="width:70%"/></center>

__Exercice 1 -__ Représentez les 3 classes dans un diagramme de classes UML. Plusieurs outils, notamment en ligne, peuvent être utilisés pour cela, par exemple [Visual paradigm](https://online.visual-paradigm.com/fr/) ou encore [diagrams.net](https://app.diagrams.net). Pour vos diagrammes, il est recommandé de commencer les noms des classes par une majuscule et les attributs par une minuscule. Durant tout ce TD, on considérera uniquement des attributs privés.

__Exercice 2 -__ Mettez à jour le diagramme UML en incluant la classe __Forme__ et les relations d'héritage. Seuls les attributs seront inclus pour le moment.

__Exercice 3 -__ On vous demande de supporter a minima pour chaque forme les méthodes suivantes :

* `translation(dx, dy)`, qui effectue une translation selon un vecteur donné.
* `contient_point(x, y)`, qui renvoie `True` si et seulement si le point donné est à l'intérieur de la forme ou sur sa frontière.
* `redimension_par_points(x0, y0, x1, y1)`, qui redimensionne la forme telle qu'elle soit incluse dans le rectangle de coins (`x0`, `y0`) et (`x1`, `y1`). Dans le cas du cercle, il faudra également qu'il soit le plus proche du premier coin. Cette méthode est utile par exemple dans [diagrams.net](https://app.diagrams.net) pour le tracé de formes par appui-déplacement de souris.

Complétez le diagramme UML avec ces méthodes. Les constructeurs devront également être renseignés (méthode `__init__` en _Python_), ainsi que les méthodes d'accès (les _getter_/_setter_) et de conversion en chaîne de caractères (méthode `__str__`). Cette méthode `__str__` permet d'obtenir la représentation d'un objet sous forme d'une chaîne de caractères, ce qui est utilisé notamment par la fonction `print`.


__Exercice 4 -__ Écrivez un squelette de code correspondant à votre diagramme UML, dans un fichier _formes.py_. Seuls les constructeurs devront être implémentés. À l'intérieur des autres méthodes, vous mettrez l'instruction `pass` de _Python_ (qui ne fait rien mais vous rappelle que le code est inachevé).


---
## Implémentation des méthodes (2h30)

Créez un fichier _test_formes.py_ dans le même dossier que _formes.py_ et initialisé avec le code suivant :

```python
from formes import *

def test_Rectangle():
    r = Rectangle(10, 20, 100, 50)
    str(r)
    assert r.contient_point(50, 50)
    assert not r.contient_point(0, 0)
    r.redimension_par_points(100, 200, 1100, 700)
    assert r.contient_point(500, 500)
    assert not r.contient_point(50, 50)

def test_Ellipse():
    e = Ellipse(60, 45, 50, 25)
    str(e)
    assert e.contient_point(50, 50)
    assert not e.contient_point(11, 21)
    e.redimension_par_points(100, 200, 1100, 700)
    assert e.contient_point(500, 500)
    assert not e.contient_point(101, 201)

def test_Cercle():
    c = Cercle(10, 20, 30)
    str(c)
    assert c.contient_point(0, 0)
    assert not c.contient_point(-19, -9)
    c.redimension_par_points(100, 200, 1100, 700)
    assert c.contient_point(500, 500)
    assert not c.contient_point(599, 500)

if __name__ == '__main__':
    test_Rectangle()
    test_Ellipse()
    test_Cercle()
```

La commande `assert` de _Python_ permet de spécifier une assertion (une condition qui doit toujours être vraie) à un point du programme. Elle sert, avant un bloc de code, à en documenter les prérequis et, après un bloc de code, à en vérifier les résultats. Son échec signifie alors un bug du programme. `assert` reçoit une expression (comme ce qu'on passe à `if`), et vérifie son résultat :

* Si `True`, l'assertion est vraie donc pas de problème, `assert` ne fait rien.
* Si `False`, l'assertion est fausse donc une exception `AssertionError` est déclenchée.
* Si l'expression renvoie une autre valeur, celle-ci est convertie en booléen pour se ramener aux deux cas précédents.

La vérification de cette condition est faite une fois au moment de son exécution (l'assertion ne sera pas valide dans le reste du programme). Dans _test_formes.py_, on utilise `assert` pour tester une fonctionnalité qui n'est pas encore implémentée, l'exécution de ce fichier échouera tant que les méthodes de seront pas codées. À l'issue de cette partie, elle ne devra renvoyer plus aucune erreur !

__Exercice 5 -__ Implémentez les méthodes de conversion en chaîne de caractères `__str__` (notamment utilisées par la fonction `print`) de chacune des classes dans _formes.py_. Vous pourrez vérifier leur bon fonctionnement en exécutant _formes.py_ (bouton `Run File - F5`), puis par exemple avec une commande `print(Rectangle(0, 0, 10, 10))` dans la console _IPython_.

__Exercice 6 -__ Implémentez les méthodes d'accès (_getter_/_setter_) pour les champs privés de chacune des classes. Pour vérifier que les champs sont bien privés, le code suivant __doit__ échouer avec une erreur de type `AttributeError` :

```python
r = Rectangle(0, 0, 10, 10)
print(r.__x, r.__y, r.__l, r.__h)
```

__Exercice 7 -__ Implémentez les méthodes `contient_point` des deux sous-classes. Vous vérifierez que les deux premiers `assert` des méthodes de test ne déclenchent pas d'erreur.

__Exercice 8 -__ Implémentez les méthodes `redimension_par_points` de chacune des sous-classes. Le fichier _test_formes.py_ doit à présent s'exécuter sans problème.

__Exercice 9 -__ (Bonus) Pour aller plus loin : on considère dans cette partie une classe __Dessin__ comme une agrégation d’objets de la classe __Forme__ :
* Un objet de la classe __Dessin__ « contient » 0, 1 ou plusieurs objets de la classe __Forme__.
* Un objet de la classe __Forme__ peut être inclus dans 0, 1 ou plusieurs objets de la classe __Dessin__.

Les dessins seront identifiés par leur nom. La classe __Dessin__ devra donc comporter un attribut `nom` ainsi que des méthodes permettant d’ajouter des formes, d’en supprimer, et d’afficher les propriétés des objets qu’elle contient.
La relation d'agrégation entre __Dessin__ et __Forme__ devant être birectionnelle, il faudra modifier la classe __Forme__ afin de permettre d'ajouter et de supprimer des dessins, ainsi que d'afficher le nom des dessins auxquels la forme est associée.

Ecrire le diagramme de classe correspondant, puis écrire le code Python associé.


