**Sommaire**

[[_TOC_]]

# TD #2 : Bibliothèque

---
## Objectif du sujet

Dans ce TD, nous abordons deux concepts fondamentaux de la programmation orientée objet (POO), __l'encapsulation__ et la __composition__, qui ont été vus lors du premier cours. Nous nous exercerons également aux [diagrammes de classes](https://fr.wikipedia.org/wiki/Diagramme_de_classes) du langage graphique [UML](https://fr.wikipedia.org/wiki/UML_(informatique)). Pour vos propres diagrammes, vous pouvez utiliser [Visual paradigm](https://online.visual-paradigm.com/fr/) ou encore [diagrams.net](https://app.diagrams.net).

Il s'agit de concevoir et de réaliser un programme _simple_ de gestion d'une bibliothèque, intégrant des lecteurs, des livres et des emprunts.

_Remarque_ : Cet énoncé part d'un problème simple et connu qui permet d'en faire la conception et la réalisation dans le temps qui nous est imparti. Les choix de conception et de réalisation sont donc orientés par ces contraintes et par les objectifs pédagogiques, à savoir : apprendre la POO en _Python_. Il est clair que le même problème dans un cadre professionnel serait traité d’une autre manière et une solution basée sur des bases de données émergerait naturellement, solution que nous écartons a priori car en dehors du périmètre de ce cours.

---
## Cahier des charges

Le cahier des charges de notre application est décrit ci-dessous. Il est volontairement donné de manière informelle.

On doit pouvoir gérer le fond documentaire d'une bibliothèque identifiée par son nom. 

1. Notre application doit être capable de gérer des lecteurs. Chacun d’eux est caractérisé par :

     - Son nom,
     - Son prénom,
     - Son adresse,
     - Un numéro (entier positif attribué de manière unique par les bibliothécaires).

1. Pour simplifier, on considérera que tous les ouvrages sont des livres, caractérisés par :

     - Le nom de l’auteur,
     - Le titre de l’ouvrage,
     - Un numéro de livre (attribué de manière unique par les bibliothécaires),
     - Le nombre d’exemplaires achetés.

1. On doit pouvoir ajouter un lecteur à une bibliothèque à tout moment et rechercher un lecteur (par son numéro de lecteur et son nom). On doit également pouvoir ajouter des ouvrages (acquisition de livres), et rechercher un ouvrage (par son numéro de livre et son titre).   
1. Tout lecteur peut emprunter des livres à la bibliothèques. Un lecteur peut emprunter plusieurs livres différents simultanément ou à des dates différentes et un même livre peut être emprunté par plusieurs lecteurs (s’il existe plusieurs exemplaires). Au moment de l’emprunt, il faut donc vérifier qu’un exemplaire de l'ouvrage est bien disponible et qu'il n'a pas été emprunté par le même lecteur. De manière symétrique, il faut également gérer le retour des livres.

1. On désire également pouvoir visualiser des états détaillés :

    - Liste de tous les lecteurs d'une bibliothèque
    - Liste de tous les livres d'une bibliothèque
    - Liste de tous les emprunts d'une bibliothèque

1. On souhaite pouvoir retirer un lecteur (s'il n'a plus d'emprunt en cours), et retirer un exemplaire non emprunté d'un livre (désherbage ou vol). Un livre qui n'aurait plus d’exemplaire ne doit plus apparaître dans la liste des livres à disposition de la bibliothèque. 

_Remarque_ : Même si ce n'est pas pas obligatoire, il vous est demandé de développer chaque classe dans un fichier _Python_ séparé (pensez à enregistrer tous les fichiers dans le même répertoire).


---
## Modélisation UML (1 heure)

__Exercice 1 -__ Il s'agit de faire la conception de cette application en utilisant une approche objet. Dans ce cadre, il faut commencer par identifier les _concepts_ qui sont manipulés par l'application. On en déduit ainsi les objets manipulés par l'application qu’on généralise en terme de classes avec les attributs et les méthodes associés.
Il est donc demandé, à partir du cahier des charges, de faire la conception de cette application en utilisant la notation UML (diagramme de classes seulement). Il faudra alors faire apparaître clairement les classes, les relations entre classes et détailler les attributs et les méthodes de chaque classe.

---
## Mise en oeuvre en Python (3 heures)

Nous allons maintenant développer en _Python_ les classes abordées précédemment. Il est vivement conseillé de faire un travail incrémental en ne développant que pas à pas chaque classe et en testant le code par des petits jeux d’essai (tests unitaires).

Pour chaque classe, on peut par exemple prévoir deux fichiers :
- Un fichier de définition de la classe (par exemple _lecteur.py_)
- Un fichier de test unitaire de la classe contenant un programme principal dont le rôle sera de tester toutes les fonctionnalités de la classe dans les différents contextes qui peuvent se présenter (par exemple _test_lecteur.py_)

L'ordre de développement des classes conseillé est le suivant : classe __Personne__, classe __Lecteur__, classe __Livre__, classe __Emprunt__, et finalement classe __Bibliotheque__.

__Exercice 2 : classe Personne -__ Ecrire le code de la classe __Personne__ dans un fichier _personne.py_. Vérifier ensuite le bon fonctionnement de cette classe en exécutant le code suivant préalablement copié dans un fichier _test_personne.py_.

``` python
# fichier test_personne.py
from personne import *

p = Personne("Durand","Marie","Ecully")

print(p)

p.set_nom("Dupond")
print(p.get_nom())
  
p.set_prenom("Emilie")
print(p.get_prenom())
    
p.set_adresse("Lyon")
print(p.get_adresse())    

print(p)
```

__Exercice 3 : classe Lecteur -__ Ecrire le code de la classe __Lecteur__ dans un fichier _lecteur.py_. Vérifier ensuite le bon fonctionnement de cette classe en exécutant le code suivant préalablement copié dans un fichier _test_lecteur.py_.

``` python
# fichier test_lecteur.py
from lecteur import *

l = Lecteur("Durand","Marie","Ecully",13)

print(l)

l.set_nom("Dupond")
print(l.get_nom())

l.set_prenom("Emilie")
print(l.get_prenom())
    
l.set_adresse("Lyon")
print(l.get_adresse())   

l.set_numero(14)
print(l.get_numero())
        
l.set_nb_emprunts(2)
print(l.get_nb_emprunts())

print(l)
```

__Exercice 4 : classe Livre -__ Ecrire le code de la classe __Livre__ dans un fichier _livre.py_. Vérifier ensuite le bon fonctionnement de cette classe en exécutant le code suivant préalablement copié dans un fichier _test_livre.py_.

``` python
# fichier test_livre.py
from livre import *

l = Livre('Le Pere Goriot','Honore de Balzac',101,2)
print(l)

l.set_auteur("Emilie Bronte")
print(l.get_auteur())
    
l.set_titre("Les Hauts de Hurlevent")
print(l.get_titre())
    
l.set_numero(102)
print(l.get_numero())

l.set_nb_total(5)
print(l.get_nb_total())

l.set_nb_dispo(4)
print(l.get_nb_dispo())

print(l)
```

__Exercice 5 : classe Emprunt -__ Ecrire le code de la classe __Emprunt__ dans un fichier _emprunt.py_. Vérifier ensuite le bon fonctionnement de cette classe en exécutant le code suivant préalablement copié dans un fichier _test_emprunt.py_.

_Remarque_ : Pour la date figurant dans cette classe, on pourra utiliser l’instruction `date.isoformat(date.today())`, en ayant pris soin d'importer la librairie : `from datetime import date` (à positionner tout en haut du fichier).


``` python
# fichier test_emprunt.py
from emprunt import *

e = Emprunt(3,5)
print(e)

print(e.get_numero_lecteur())
print(e.get_numero_livre())
print(e.get_date())
```

__Exercice 6 : classe Bibliotheque -__ Ecrire finalement le code de la classe __Bibliotheque__ dans un fichier _bibliotheque.py_. Vérifier ensuite le bon fonctionnement de cette classe en exécutant le code suivant préalablement copié dans un fichier _test_bibliotheque.py_.


``` python
from bibliotheque import *

# Creation d'une bibliotheque
b = Bibliotheque('Bibliotheque ECL')

# Ajout de lecteurs
b.ajout_lecteur('Duval','Pierre','rue de la Paix',1)
b.ajout_lecteur('Dupond','Laurent','rue de la Gare',2)
b.ajout_lecteur('Martin','Marie','rue La Fayette',3)
b.ajout_lecteur('Dubois','Sophie','rue du Stade',4)

# Ajout de livres
b.ajout_livre('Le Pere Goriot','Honore de Balzac',101,2)
b.ajout_livre('Les Hauts de Hurlevent','Emilie Bronte',102,2)
b.ajout_livre('Le Petit Prince','Antoine de Saint Exupery',103,2)
b.ajout_livre('L\'Etranger','Albert Camus',104,2)

# Affichage des lecteurs et des livres
print('\n--- Liste des lecteurs :')
print('-------------------------------')
b.affiche_lecteurs()
print('\n--- Liste des livres :')
print('-------------------------------')
b.affiche_livres()

# Recherches de lecteurs par numero
print('\n--- Recherche de lecteurs :')
print('-------------------------------')
lect = b.chercher_lecteur_numero(1)
if lect != None:
    print(lect)
else:
    print('Lecteur non trouve')

lect = b.chercher_lecteur_numero(6)
if lect != None:
    print(lect)
else:
    print('Lecteur non trouve')

# Recherches de lecteurs par nom
lect = b.chercher_lecteur_nom('Martin','Marie')
if lect != None:
    print(lect)
else:
    print('Lecteur non trouve')
    
lect = b.chercher_lecteur_nom('Le Grand','Paul')
if lect != None:
    print(lect)
else:
    print('Lecteur non trouve')

# Recherches de livres par numero
print('\n--- Recherche de livres :')
print('-------------------------------')
livre = b.chercher_livre_numero(101)
if livre != None:
    print('Livre trouve :',livre)
else:
    print('Livre non trouve')

livre = b.chercher_livre_numero(106)
if livre != None:
    print('Livre trouve :',livre)
else:
    print('Livre non trouve')

# Recherches de livres par titre
livre = b.chercher_livre_titre('Les Hauts de Hurlevent')
if livre != None:
    print('Livre trouve :',livre)
else:
    print('Livre non trouve')

livre = b.chercher_livre_titre('Madame Bovarie')
if livre != None:
    print('Livre trouve :',livre)
else:
    print('Livre non trouve')

# Quelques emprunts
print('\n--- Quelques emprunts :')
print('-------------------------------')
b.emprunt_livre(1,101)
b.emprunt_livre(1,104)
b.emprunt_livre(2,101)
b.emprunt_livre(2,105)
b.emprunt_livre(3,101)
b.emprunt_livre(3,104)
b.emprunt_livre(4,102)
b.emprunt_livre(4,103)

# Affichage des emprunts, des lecteurs et des livres
print('\n--- Liste des emprunts :')
print('-------------------------------')
b.affiche_emprunts()
print('\n--- Liste des lecteurs :')
print('-------------------------------')
b.affiche_lecteurs()
print('\n--- Liste des livres :')
print('-------------------------------')
b.affiche_livres()

# Quelques retours de livres
print('\n--- Quelques retours de livres :')
print('-------------------------------')
b.retour_livre(1,101)
b.retour_livre(1,102)
b.retour_livre(3,104)
b.retour_livre(10,108)

# Affichage des emprunts, des lecteurs et des livres
print('\n--- Liste des emprunts :')
print('-------------------------------')
b.affiche_emprunts()
print('\n--- Liste des lecteurs :')
print('-------------------------------')
b.affiche_lecteurs()
print('\n--- Liste des livres :')
print('-------------------------------')
b.affiche_livres()

# Suppression de quelques livres
rep = b.retrait_livre(101)
if not rep:
    print('Retrait du livre impossible')
else:
    print('Retrait du livre effectue')

b.retour_livre(2,101)

rep = b.retrait_livre(101)
if not rep:
    print('Retrait du livre impossible')
else:
    print('Retrait du livre effectue')

# Suppression de quelques lecteurs
rep = b.retrait_lecteur(1)
if not rep:
    print('Retrait du lecteur impossible')
else:
    print('Retrait du lecteur effectue')

b.retour_livre(1,104)

rep = b.retrait_lecteur(1)
if not rep:
    print('Retrait du lecteur impossible')
else:
    print('Retrait du lecteur effectue')

# Affichage des emprunts, des lecteurs et des livres
print('\n--- Liste des emprunts :')
print('-------------------------------')
b.affiche_emprunts()
print('\n--- Liste des lecteurs :')
print('-------------------------------')
b.affiche_lecteurs()
print('\n--- Liste des livres :')
print('-------------------------------')
b.affiche_livres()
```