**Sommaire**

[[_TOC_]]


# Tutoriel sur git/gitlab

Ce document présente un tutoriel à l'usage de _git_/_gitlab_. Il s'appuie sur celui proposé par __Stéphane Derrode__.

L'objectif est de découvrir _git_ et de savoir l'utiliser dans un contexte simple (un seul utilisateur, une seule branche de développement). Un autre tutoriel s'intéressera à des fonctionnalités plus avancées.

Avant toute chose, afin d'utiliser _git_, il est nécessaire de savoir ouvrir un _Terminal_, quel que soit le système d'exploitation (_Windows_, _Linux_, ou _Mac OS X_). Sous _Windows X_, vous pourrez utiliser le programme _Windows powershell_ qui est très similaires au _Terminal_ de _Linux_ et de _Mac OS X_. Vous devez aussi savoir naviguer dans vos dossiers à l'aide de la commande ```cd```. Typiquement :

```shell
cd c:\Users\toto\monTD  # Sous Windows
cd ~/monTD              # Sous Mac, linux
cd ..                   # Pour remonter d'un niveau dans la hiérarchie des dossiers.
```

---
## Installer _git_ pour Windows (si requis) et *Github Desktop*

Tout d'abord, il faut vérifier que _git_ (un gestionnaire de version de fichiers) est bien installé sur votre machine. _git_ est normalement disponible par défaut sur les machines *Mac OS X*, *Linux* et *Windows 10*. Pour vérifier, lancez la commande suivante dans un _Terminal_
```shell
git --version
```
Si _git_ est absent, alors installez-le grâce au lien suivant [git-scm](https://git-scm.com/download/win). Lors de l'installation, validez les choix par défaut qui vous sont proposés.

L'interface *Github Desktop* que nous allons installer n'est absolument pas nécessaire pour travailler avec _git_. Cependant, elle est pratique dans la mesure où elle évite d'avoir à utiliser des commandes manuelles à partir d'un _Terminal_.

 1. Télécharger, installer et exécuter l'application [Github Desktop](https://desktop.github.com/).
 <center><img src="figures/GithubDesktop_vierge.png" style="width:50%"/></center>
 
 2. Configurer *Github Desktop* (cette étape n'est nécessaire que si cela n'a pas déjà été fait lors de l'installation du logiciel) : aller dans le menu ``Préférences / git`` et entrez le nom ainsi que l'adresse électronique.


---
## Gérer son premier projet _git_ avec _gitlab_

 1. Accéder au _gitlab_ de l'Ecole Centrale :    
    - Adresse : [https://gitlab.ec-lyon.fr](https://gitlab.ec-lyon.fr);    
    - Identifiants : ceux de l'ECL;     
    - Naviguer dans le projet INF-TC2.
 1. Créer un nouveau projet :    
    - Nom : _HelloWorld_;     
    - Description : Mon premier projet _gitlab_;    
    - Choisir parmi *Public/Internal/Private*;   
    - Sélectionner *Initialize repository with a README*;
    <center><img src="figures/Gitlab_newproject.png" style="width:50%"/></center>
    - Cliquer sur l'icône `Clone` et copier le lien `Clone with HTTPS`, qui ressemble à `https://gitlab.ec-lyon.fr/xyyyyy/helloworld.git`. 
 1. Basculer sur *Github Desktop*,     
    - Sélectionner l'option `Clone a Repository from the Internet`, puis l'onglet `URL` (ou menu `File > Clone repository`).
    - Recopier l'adresse précédemment mise en mémoire dans l'espace dédié, et choisir dans `local path` le répertoire dans lequel votre projet sera copié. Appuyez sur le bouton `Clone` de l'interface pour lancer l'importation de votre projet _gitlab_ sur votre machine.
    - Vérifier avec un gestionnaire de fichiers que le *repo* est bien copié localement sur sa machine et qu'il contient un répertoire caché (`.git`) ainsi que le fichier *readme.md*.   
    - Basculer sur l'onglet `history`, et commenter.     
 1. Pour éviter que certains fichiers temporaires ne soient suivis par _git_, on va spécifier les fichiers a exclure dans un fichier appelé *.gitignore* (il s'agit d'un fichier caché). Pour cela    
    - Aller dans le menu `Repository > Repository settings`, puis `Ignored Files`.   
    - Copier le contenu du fichier qui se trouve à l'adresse [https://github.com/github/gitignore/blob/master/Python.gitignore](https://github.com/github/gitignore/blob/master/Python.gitignore) dans l'espace réservé, puis `Save`. <center><img src="figures/gitignore.png" style="width:50%"/></center>     
        
    - On doit alors *commiter* les changements (le texte par défaut qui décrit le `commit` est OK). Plus de détails sur ces `commits` plus loin.      
 1. Éditer le fichier *readme.md* avec *Visual Code* (ou tout autre éditeur de texte) et enregistrer les changements localement :
    - Basculer vers *Github Desktop* et observer les changements qui interviennent sur l'interface.    
    - *Commiter* les changements, (visibles ensuite dans l'onglet *History*)    
    - *Pusher* le travail vers _gitlab_ en utilisant l'option *Publish Branch*.
    - Basculer vers _gitlab_ et vérifier que les changements ont bien étaient publiés   
 1. Basculer vers *Spyder* pour éditer du code _Python_ :   
    - Taper un programme *incremente_de_un(x)* et sauvegarder dans votre répertoire local sous le nom de fichier *incremente.py*. 
    - *commiter* les changements dans *Github Desktop*   
    - Refaire la même chose avec un autre fichier (*incremente_de_deux(x)*), à sauvegarder dans un sous répertoire *test*.
    - *Pusher* son travail vers _gitlab_, et vérifier que tout se met bien à jour.


De manière synthétique, une fois le projet créé,

   1. vous modifiez vos code sources, et enregistrez régulièrement vos changements avec un `commit` (donc en local). Vous pouvez `commiter` des sources même s'ils ne sont pas terminés ou comportent des bugs. À ce stade cela n'a pas d'importance.     
   1. quand vous êtes content de votre travail, et que vous jugez vos algorithmes fonctionnels, vous pouvez alors publier vos précédents `commit` sur _gitlab_ avec un `push`.

Ce comportement est illustré par le schéma de la figure suivante:
<center><img src="figures/PrincipeGit.png" style="width:50%"/></center>   

-----
Pour détruire ce projet, il faut :

 1. Quitter Spyder pour fermer les fichiers en cours.
 1. Détruire la version locale en utilisant *Github Desktop*: sélectionner l'onglet `Current Repository`, et le menu contextuel sur le nom de votre projet permet d'accéder à `Remove`. Penser à cocher la suppression physique des fichiers si vous le souhaitez (sinon seul la gestion du projet par *Github Desktop* sera supprimée, les fichiers seront toujours présents sur votre disque dur).
 1. Détruire la version distante sur _gitlab_: Dans la barre de menu latérale à gauche de l'interface Web, sélectionnez l'icône représentant une roue crantée correspondant à `Settings`. Dans le sous-menu `Général`, appuyez sur le bouton `Expand` correspondant à la ligne : *Advanced*. Tout en bas, sélectionner `Delete Project`, copier la phrase demandée et appuyer sur le bouton pour confirmer.


-----
A quoi ça sert d'utiliser _git_ et _gitlab_ ? Les avantages sont nombreux :

 - Avoir une copie de son travail sur un serveur distant.
 - Réaliser des sauvegardes incrémentales et disposer d'un historique des modifications (undo infini!)
 - Partager son travail avec tout le monde (si votre projet est publique), ou avec des personnes choisies (invitations personnelles).
 - Travailler à plusieurs sur le même projet et en même temps    
      - fusion automatique (ou presque) des sources,
      - traçage des modification (qui, quoi).
      - ...
